##This script is meant to calculate shanten number for a  mahjong hand of 13 tiles
##It does not count shanten for chiitoitsu or kokushi

##Intro:
##A hand in mahjong contains 13 tiles
##To win a game you need to gather 4 sets of tiles and a head pair in your hand
##A set can be 3 consecutive tiles in one suit or 3 identical tiles (kans are not considered in this script)
##You can get this hand structure only on your turn when you draw the 14th tile from the wall(deck of tiles)
##If you need only this 14th tile to win, you are tempai
##If you are 1 tile away from tempai — you are 1-shanten
##Shanten number shows how many tiles away from tempai you are
##Shanten can't be greater than 8

import itertools, random, copy

##There exist 3 tile suits: man, pim, sou — and 7 types of honor tiles
global tiles
tiles = list(itertools.product(['man','pin','sou'],range(1,10))) + list(itertools.product(['ton','nan','sha','pei','chun','haku','hacu'],[0]))

##To operate with tiles more efficiently, we need to encode them as ints
##1-9 are man, 21-29 are pin, 41-49 are sou
global code
nums = list(range(1, 10)) + list(range(21, 30)) + list(range(41, 50)) + [70,73,76,79,82,85,88]
code1 = {tiles[i-1]:nums[i-1] for i in range(0, 35)}
code2 = {nums[i-1]:tiles[i-1] for i in range(0, 35)}
code = code1.copy()
code.update(code2)

##make tuple hand into int hand and vice versa
def encode(hand):
    en_hand = []
    for el in hand:
        en_hand.append(code[el])
    return(sorted(en_hand))

##The 13-tile hand will be represented as a tree, where each node contains a set, a protoset, a pair or a single tile
##Each node has one parent and lots of children
##Later we will construct this tree in a top-bottom manner
class node:

    def __init__(self, value):
        self.value = value
        self.parent = None
        self.child=[]
        ##self.pair is indicator for the head pair in the hand
        ##it is different from other pairs as it does not count as protoset and there can be only one head pair in a hand
        self.pair = False
        self.protosets = 0
        ##self.sh is the shanten number
        ##shanten is the number of tiles needed to obtain tempai
        self.sh = None
        
    def add(self,other):
        other.parent = self
        self.child.append(other)

##SETS AND PROTOSETS
##Chi — 3 consecutive tiles: 1-2-3; 44-45-46 [2-shanten contribution]
##Pon — 3 identical tiles: 22-22-22, 82-82-82 [2-shanten contribution]
##Head pair — 2 identical tiles, can be only one in a hand but is required to win [1-shanten contribution]
##Sidewait — 1-2; 2-3; 45-56 [1-shanten contribution]
##Kanchan (hole-wait) — 1-.-3; 44-.-46 [1-shanten contribution]
##Pair-wait — pon protoset [1-shanten contribution]
##Single — detached tile [0-shanten contribution]
        
##(Isolated) finds such tiles that can be a part neither of chi nor pon nor head pair
##These tiles are singular in a hand and have no neighbors within 2-tile distance
def isolated(hand):
    iso = []
    temp = hand.copy()
    for el in hand:
        if (el-1 not in hand) and (el+1 not in hand) and (el-2 not in hand) and (el+2 not in hand) and (hand.count(el) == 1):
            iso.append(el)
            temp.remove(el)
    iso = tuple(iso)
##(Isolated)'s output first element is a tuple of isolated tiles. Its 2nd element is the initial hand without isolated tiles
    return(iso,temp)

def head(hand,root):
##We can make a node with head-pair only if we didn't make such a node earlier.
##If we did, we would see that (root.pair) is True
    if not (root.pair):
##We need to check the number of tiles is no less than 2
        if len(hand) < 2:
            return(None)
        a = hand[0]
        if (a == hand[1]):
            temp = (a,a)
            node2 = node(temp)
            node2.sh = root.sh-1
            node2.pair = True
            return(node2)
    return(None)

##proto-pon
def pair(hand, root):
    if len(hand) < 2:
        return(None)
    a = hand[0]
    if (a == hand[1]):
        node7 = node((a,a))
##Protosets do not decrease shanten number if there are 4+ of them, as we need exactly 4 of them in a hand to win
        if root.protosets >= 4:
            node7.sh = root.sh
            node7.protoset = root.protosets + 1
        else:
            node7.sh = root.sh - 1
            node7.protoset = root.protosets + 1
        return(node7)
    return(None)

def sidewait(hand,root):
    if len(hand) < 2:
            return(None)
    a = hand[0]
    if (a+1 in hand):
        node3 = node((a,a+1))
        if root.protosets > 3:
            node3.sh = root.sh
            node3.protoset = root.protosets + 1
        else:
            node3.sh = root.sh - 1
            node3.protoset = root.protosets + 1
        return(node3)
    return(None)

def kanchan(hand,root):
    if len(hand) < 2:
            return(None)
    a = hand[0]
    if (a+2 in hand) and (a%10 != 9):
        node4 = node((a,a+2))
        if root.protosets > 3:
            node4.sh = root.sh
            node4.protoset = root.protosets + 1
        else:
            node4.sh = root.sh - 1
            node4.protoset = root.protosets + 1
        return(node4)
    return(None)

def pon(hand,root):
    if len(hand) < 3:
            return(None)
    a = hand[0]
    if (a == hand[1]) and (a == hand[2]):
        node5 = node((a,a,a))
        node5.sh = root.sh - 2
        return(node5)
    return(None)

def chi(hand,root):
    if len(hand) < 3:
            return(None)
    a = hand[0]
    if (a+1 in hand) and (a+2 in hand):
        node6 = node((a,a+1,a+2))
        node6.sh = root.sh - 2
        return(node6)
    return(None)

def single(hand,root):
##There can't be two consecutive nodes with single tiles
##as we will make sure all the isolated tiles are separated aforehand
    if len(hand) < 1 or len(root.value) == 1:
        return(None)
##Make sure el won't become isolated if hand[0] is considered single
    el = hand[1]
    if (el+1 not in hand) and (el+2 not in hand) and (el not in hand[2:]):
        return(None)
    node1 = node((hand[0],))
    ##Single tiles do not decrease shanten-number
    node1.sh = root.sh
    return(node1)
##(Subtract) takes in tiles in hand and information about previously calculated protosets, pairs and shanten from root.
##It returns a list of nodes whose values contain found protosets (or head-pair or single tile)
def subtract(hand, root):
    nodes = [single(hand,root), head(hand,root), sidewait(hand,root), kanchan(hand,root), pair(hand,root), pon(hand,root), chi(hand,root)]
    nodes = list(filter(lambda a: a != None, nodes))
    return(nodes)

##The ultimate function to calculate shanten
def shanten(hand,root = None,num = 8):
##When root is None (first level) we create root
    if root is None:
        temp = isolated(hand)
        root = node(temp[0])
        hand = temp[1]
        root.sh = num
    nodes_to_add_to_root = subtract(hand,root)
    for i in nodes_to_add_to_root:
        hand1 = hand.copy()
        for j in i.value:
            hand1.remove(j)
        if len(hand1) == 0:
            return(i)
        a = shanten(hand1,i,i.sh)
##How come I even need this line?
##Why would (i) be the same as (a)?
        if not(i == a):
            i.add(a)
        root.add(i)
    return(root)

wall = {tiles[i-1]: 4 for i in range(0, 35)}
hand = []
L = 0
while L != 14:
    tile = random.choice(list(wall.keys()))
    if wall[tile] != 0:
        hand.append(tile)
        wall[tile] = wall[tile]-1
        L += 1
hand = [2, 4, 4, 7, 26, 27, 28, 43, 44, 49, 49, 85, 88, 88]
a = shanten(hand[:-1])
print(a.value)
print(hand[:-1])
print('a.child.values')
for el in a.child:
    print(el.value)